package rangesearch;

import rangesearch.api.*;
import rangesearch.ref.*;

import java.io.*;
import jdsl.core.api.*;
import jdsl.core.ref.*;
import jdsl.geomobj.api.*;
import jdsl.geomobj.ref.*;
import javax.swing.event.EventListenerList;

/** A representation of the GUI document structure, including a set of points
 * and a range.*/

public class RangeModel {
    protected Sequence m_points;
    protected Rectangle2D m_range;
    protected EventListenerList m_listenerList = new EventListenerList();

    public RangeModel() {
        m_points = new ArraySequence();
        m_range  = null;
    }

    public void generatePoints(int numPoints, int x1, int y1, int x2, int y2) {
        int w = x2 - x1;
        int h = y2 - y1;
        m_points = new ArraySequence();
        for(int i = 0; i < numPoints; i ++) {
            int x = x1 + (int) (Math.random() * w);
            int y = y1 + (int) (Math.random() * h);
            m_points.insertLast(new IntPoint2D(x, y));
        }
        fireModelChanged(ModelChangedEvent.POINTS_CHANGED);
    }

    /** Sets the model's set of points. */
    public void setPoints(InspectableSequence points) {
        m_points = new ArraySequence();
        ObjectIterator oi = points.elements();
        while(oi.hasNext()) {
            m_points.insertLast((Point2D) oi.nextObject());
        }
        fireModelChanged(ModelChangedEvent.POINTS_CHANGED);
    }

    /** Adds a point to the model.
     * @return the index of the new point within the sequence of points. */
    public int addPoint(Point2D p) {
        m_points.insertLast(p);
        fireModelChanged(ModelChangedEvent.POINTS_CHANGED);
        return (m_points.size() - 1);
    }

    /** Sets the point at the given index. */
    public void setPoint(int index, Point2D p) {
        m_points.removeAtRank(index);
        m_points.insertAtRank(index, p);
        fireModelChanged(ModelChangedEvent.POINTS_CHANGED);
    }

    /** Removes a point from the model. */
    public void removePoint(int index) {
        m_points.removeAtRank(index);
        fireModelChanged(ModelChangedEvent.POINTS_CHANGED);
    }

    /** @return the point at the given index. */
    public Point2D getPoint(int index) {
        return (Point2D) m_points.atRank(index).element();
    }

    public Point2DIterator getPoints() {
        return new Point2DIteratorAdapter(m_points.elements());
    }

    /** @return the number of points in the model. */
    public int getNumPoints() {
        return m_points.size();
    }

    /** Sets the model's range. */
    public void setRange(Rectangle2D range) {
        m_range = range;
        fireModelChanged(ModelChangedEvent.RANGE_CHANGED);
    }

    /** Returns the model's range. */
    public Rectangle2D getRange() {
        return m_range;
    }

    /** Adds a listener for the ModelChangedEvent. */
    public void addModelChangedListener(ModelChangedListener l) {
        m_listenerList.add(ModelChangedListener.class, l);
    }

    /** Removes a listener for the ModelChangedEvent. */
    public void removeModelChangedListener(ModelChangedListener l) {
        m_listenerList.remove(ModelChangedListener.class, l);
    }

    /** Notifies all listeners that the current tool has changed. */
    protected void fireModelChanged(int changed) {
        // Guaranteed to return a non-null array
        Object[] listeners = m_listenerList.getListenerList();
        // Process the listeners last to first, notifying
        // those that are interested in this event
        for(int i = listeners.length - 2; i >= 0; i -= 2) {
            if(listeners[i] == ModelChangedListener.class) {
                // Lazily create the event:
                ModelChangedEvent e = new ModelChangedEvent(this, changed);
                ((ModelChangedListener)listeners[i+1]).modelChanged(e);
            }
        }
    }

    /** Writes the model to a text stream. */
    public void save(Writer s)
            throws IOException {
        s.write(m_points.size() + "\n");
        for(int i = 0; i < m_points.size(); i ++) {
            Point2D p = (Point2D) m_points.atRank(i).element();
            JDSLUtil.savePoint(s, p);
        }
        Point2D p1 = null;
        Point2D p2 = null;
        if(m_range != null) {
            p1 = m_range.topRightPoint();
            p2 = m_range.bottomLeftPoint();
        }
        JDSLUtil.savePoint(s, p1);
        JDSLUtil.savePoint(s, p2);
    }

    /** Reads the model from a text stream. */
    public void load(StreamTokenizer st)
            throws IOException {
        ArraySequence points = new ArraySequence();
        int numPoints = (int) JDSLUtil.readNumber(st);
        for(int i = 0; i < numPoints; i ++) {
            Point2D p = JDSLUtil.loadPoint(st);
            points.insertLast(p);
        }
        Point2D p1 = JDSLUtil.loadPoint(st);
        Point2D p2 = JDSLUtil.loadPoint(st);

        m_points = points;
        m_range = new RatRectangle2D(p1, p2);
        fireModelChanged(ModelChangedEvent.ALL_CHANGED);
    }
}


