package rangesearch.ref;

import rangesearch.api.*;
import rangesearch.JDSLUtil;

import jdsl.core.algo.sorts.ArrayQuickSort;
import jdsl.core.api.*;
import jdsl.core.ref.*;
import jdsl.geomobj.api.*;
import jdsl.geomobj.ref.*;

/** An implementation of a two-dimensional range tree, using a balanced binary
 * tree as the primary structure and an array as the secondary structure. */

public class RangeTree2DImpl implements RangeTree2D {
    protected Comparator m_comp1;
    protected Comparator m_comp2;
    protected BinaryTree m_tree;

    public RangeTree2DImpl() {
        m_tree = null;
        m_comp1 = null;
        m_comp2 = null;
    }

    /** Initializes the range tree to the given set of elements, with the given
     * comparator.
     * @param oi    an iterator over the elements to search.
     * @param comp1 the comparator for the primary structure (used for building
     *              the binary tree).
     * @param comp2 the comparator for the secondary structure (used for
     *              sorting the array.
     */
    public void initialize(ObjectIterator oi,
                           Comparator comp1,
                           Comparator comp2) {
        m_comp1 = comp1;
        m_comp2 = comp2;

        // Collect the elements into an array and sort by the first comparator.
        ArraySequence s = new ArraySequence();
        while(oi.hasNext()) {
            s.insertLast(oi.nextObject());
        }
        ArrayQuickSort quickSort = new ArrayQuickSort();
        quickSort.sort(s, m_comp1);

        // Construct a balanced binary tree over the elements.
        if(s.size() == 0) {
            // Empty tree is represented by a null tree.
            m_tree = null;
        } else {
            m_tree = new NodeBinaryTree();
            constructTree(m_tree.root(), s, 0, s.size());

            // Build the range sets at each internal node.
            buildRangeSet(m_tree.root());

            // Augment each node with its range boundaries.
            buildRangeBounds(m_tree.root());
        }
    }

    /** Recursively constructs a tree, expanding the given node to include the
     * given subsequence of elements.
     * @param subTreeRoot the node to expand.
     * @param s           the sequence from which to extract elements.
     * @param start       the start index of the sequence to extract.
     * @param end         the end index of the sequence to extract.
     */
    protected void constructTree(Position subTreeRoot,
                                 ArraySequence s, int start, int end) {
        int len = end - start;
        Assertion.check(len > 0);
        if(len == 1) {
            // Store a sequence at a leaf, containing just the element.
            ArraySequence leafSeq = new ArraySequence();
            Object obj = s.atRank(start).element();
            leafSeq.insertLast(obj);
            m_tree.replaceElement(subTreeRoot, leafSeq);
        } else {
            m_tree.expandExternal(subTreeRoot);
            int half = (start + end) / 2;
            constructTree(m_tree.leftChild(subTreeRoot), s, start, half);
            constructTree(m_tree.rightChild(subTreeRoot), s, half, end);
        }
    }

    /** Recursively builds the secondary structure (a sorted array) at each
     * internal node.
     * @param subTreeRoot the node for which to build a secondary structure.
     */
    protected void buildRangeSet(Position subTreeRoot) {
        if(m_tree.isExternal(subTreeRoot)) return;

        // For each internal node, build the left and right range sets.
        Position lTree = m_tree.leftChild (subTreeRoot);
        Position rTree = m_tree.rightChild(subTreeRoot);
        buildRangeSet(rTree);
        buildRangeSet(lTree);

        Sequence s1 = (Sequence) lTree.element();
        Sequence s2 = (Sequence) rTree.element();

        Sequence s = mergeSortedArrays(s1, s2);

        // Store the new range set at this node.
        m_tree.replaceElement(subTreeRoot, s);
    }

    /** Merges two sorted arrays together into another sorted array, as sorted
     * by the secondary comparator.
     * @param s1 the first array to merge.
     * @param s2 the second array to merge.
     * @return the merged, sorted sequence.
     */
    protected Sequence mergeSortedArrays(Sequence s1,
                                         Sequence s2) {
        // Merge the range sets together.
        Sequence s = new ArraySequence();
        int numElements1 = s1.size();
        int numElements2 = s2.size();
        int numElements = numElements1 + numElements2;
        int j = 0;
        int k = 0;
        for(int i = 0; i < numElements; i ++) {
            if(j >= numElements1) {
                s.insertLast(s2.atRank(k).element());
                k ++;
            } else if(k >= numElements2) {
                s.insertLast(s1.atRank(j).element());
                j ++;
            } else {
                Object obj1 = s1.atRank(j).element();
                Object obj2 = s2.atRank(k).element();
                if(m_comp2.isLessThanOrEqualTo(obj1, obj2)) {
                    s.insertLast(s1.atRank(j).element());
                    j ++;
                } else {
                    s.insertLast(s2.atRank(k).element());
                    k ++;
                }
            }
        }
        return s;
    }

    /** Extracts the minimum and maximum elements of the subtree and stores
     * them as decorations with the string tags "min" and "max".
     * @param subTreeRoot the subtree on which to operate.
     */
    protected void buildRangeBounds(Position subTreeRoot) {
        if(m_tree.isExternal(subTreeRoot)) {
            // If we're dealing with a leaf, extract its single seconary
            // element and use it as the minimum and maximum elements.
            ArraySequence leafSeq = (ArraySequence) subTreeRoot.element();
            Object obj = leafSeq.atRank(0).element();
            subTreeRoot.set("min", obj);
            subTreeRoot.set("max", obj);
        } else {
            // If we're dealing with an internal node, use the minimum from its
            // left subtree and the maximum from its right subtree.
            Position lTree = m_tree.leftChild (subTreeRoot);
            Position rTree = m_tree.rightChild(subTreeRoot);
            buildRangeBounds(lTree);
            buildRangeBounds(rTree);
            subTreeRoot.set("min", lTree.get("min"));
            subTreeRoot.set("max", rTree.get("max"));
        }
    }

    /** Queries the elements in the range tree with the given ranges.
     * Elements equal to the range bounds are considered to be within the
     * range.
     * @param start1 the starting element for the primary search range.
     * @param end1 the ending element for the primary search range.
     * @param start2 the starting element for the secondary search range.
     * @param end2 the ending element for the secondary search range.
     * @return an iterator over the elements in the given ranges.
     */
    public ObjectIterator query(Object start1, Object end1,
                                Object start2, Object end2) {
        Assertion.check(m_comp1 != null, "query called before initialize.");
        Assertion.check(m_comp2 != null, "query called before initialize.");

        // Check for empty tree.
        if(m_tree == null) {
            return (new ArraySequence()).elements();
        }

        // Find the allocation nodes that correspond to the first query range.
        Sequence allocNodes = new ArraySequence();
        findAllocNode(m_tree.root(), allocNodes, start1, end1);

        // Query each allocation node with the second query range.
        ArraySequence retObjs = new ArraySequence();
        ObjectIterator oi = allocNodes.elements();
        while(oi.hasNext()) {
            Position pos = (Position) oi.nextObject();
            ArraySequence s = (ArraySequence) pos.element();
            queryAllocNode(s, retObjs, start2, end2);
        }

        return retObjs.elements();
    }

    /** Recursively searches the binary tree for allocation nodes.
     * @param subTreeRoot the subtree to search.
     * @param sequence    the sequence to which the found allocation nodes are
     *                    appended.
     * @param start       the start of the primary search range.
     * @param end         the end of the primary search range.
     */
    protected void findAllocNode(Position subTreeRoot,
                                 Sequence allocNodes,
                                 Object start, Object end) {
        Object minObj = subTreeRoot.get("min");
        Object maxObj = subTreeRoot.get("max");

        if(m_comp1.isGreaterThanOrEqualTo(minObj, start) &&
           m_comp1.isLessThanOrEqualTo(maxObj, end)) {
            // If this node's range is entirely contained in the search range,
            // add it to the list.
            allocNodes.insertLast(subTreeRoot);
        } else if(m_tree.isInternal(subTreeRoot) &&
                  (m_comp1.isLessThanOrEqualTo(minObj, end) ||
                   m_comp1.isGreaterThanOrEqualTo(maxObj, start))) {
            // Otherwise, only recur if the node's range overlaps partially
            // with the search range.
            Position lTree = m_tree.leftChild (subTreeRoot);
            Position rTree = m_tree.rightChild(subTreeRoot);
            findAllocNode(lTree, allocNodes, start, end);
            findAllocNode(rTree, allocNodes, start, end);
        }
    }

    /**
     * Finds the nodes between start and end in the given sequence and appends
     * them to the return sequence.
     * @param s      the sequence to search.
     * @param retSeq the return sequence to which allocation nodes are append.
     * @param start  the start bounds of the secondary search range.
     * @param end    the end bounds of the secondary search range.
     */
    protected void queryAllocNode(ArraySequence s,
                                  Sequence retSeq,
                                  Object start, Object end) {
        int startIndex = binarySearchAfter(s, start, 0, s.size());
        int endIndex   = binarySearchBefore(s, end, 0, s.size());

        for(int i = startIndex; i <= endIndex; i ++) {
            retSeq.insertLast(s.atRank(i).element());
        }
    }

    /** Recursively finds the index into a sorted array of the first element
     * greater than or equal to the given element.  Returns the sequence size
     * if no element is greater than or equal to the given element.
     * @param s     the sorted array to search.
     * @param obj   the element to search for.
     * @param start the start index of the search.
     * @param end   the end index of the search.
     * @return the index of the first element greater than or equal to the
     *         given element.
     */
    protected int binarySearchAfter(ArraySequence s, Object obj,
                                    int start, int end) {
        int len = end - start;
        if(len == 0) {
            return s.size();
        } else if(len == 1) {
            if(m_comp2.isGreaterThanOrEqualTo(s.atRank(start).element(), obj)) {
                return start;
            } else {
                return start + 1;
            }
        } else {
            int half = (start + end) / 2;
            if(m_comp2.isGreaterThanOrEqualTo(s.atRank(half).element(), obj)) {
                return binarySearchAfter(s, obj, start, half);
            } else {
                return binarySearchAfter(s, obj, half, end);
            }
        }
    }

    /** Recursively finds the index into a sorted array of the last element
     * less than or equal to the given element.  Returns -1 if no element is
     * less than or equal to the given element.
     * @param s     the sorted array to search.
     * @param obj   the element to search for.
     * @param start the start index of the search.
     * @param end   the end index of the search.
     * @return the index of the last element less than or equal to the
     *         given element.
     */
    protected int binarySearchBefore(ArraySequence s, Object obj, int start, int end) {
        int len = end - start;
        if(len == 0) {
            return -1;
        } else if(len == 1) {
            if(m_comp2.isLessThanOrEqualTo(s.atRank(start).element(), obj)) {
                return start;
            } else {
                return start - 1;
            }
        } else {
            int half = (start + end) / 2;
            if(m_comp2.isLessThanOrEqualTo(s.atRank(half).element(), obj)) {
                return binarySearchBefore(s, obj, half, end);
            } else {
                return binarySearchBefore(s, obj, start, half);
            }
        }
    } 
}


